from unittest import mock

from click.testing import CliRunner

from tatt.transcribe import cli
from tatt.vendors.amazon import Transcriber


def test_services():
    runner = CliRunner()
    result = runner.invoke(cli, ['services'])
    assert result.exit_code == 0
    assert 'amazon\t\t$0.006 per 15 seconds' in result.output
    assert ('Here are all the available speech-to-text-services:' 
            in result.output)


@mock.patch('tatt.transcribe.get_transcription_jobs')
def test_status(get_transcription_jobs):
    runner = CliRunner()
    result = runner.invoke(cli, ['status', 'hi'])
    assert get_transcription_jobs.called
    get_transcription_jobs.assert_called_with(name='hi')

# list, get

@mock.patch('tatt.transcribe.get_service')
def test_this(get_service):
    runner = CliRunner()
    result = runner.invoke(cli, ['this', 'hi.mp3', 'amazon'])
    get_service.assert_called_once()
    get_service.assert_called_with('amazon')


@mock.patch('tatt.transcribe.get_transcription_jobs')
def test_list(get_transcription_jobs):
    runner = CliRunner()
    result = runner.invoke(cli, ['list'])
    get_transcription_jobs.assert_called_once()
    get_transcription_jobs.assert_called_with(None, None, None)

    result = runner.invoke(cli, ['list', '-n', 'hi.mp3'])
    get_transcription_jobs.assert_called_with(None, 'hi.mp3', None)




class ConfigError(Exception):
    pass


class AlreadyExistsError(Exception):
    pass


class DoesntExistError(Exception):
    pass


class NotAvailable(Exception):
    pass


class DependencyRequired(Exception):
    pass


class FormatError(Exception):
    pass



import io
import json
import os
import pathlib
import shutil
import tempfile
from time import sleep
from typing import List

from google.api_core import operations_v1
from google.cloud import (
    speech_v1p1beta1 as speech,
    storage,
    exceptions as gc_exceptions,
        )

from tatt import exceptions, helpers, config as config_mod
from .vendor import TranscriberBaseClass

NAME = 'google'
BUCKET_NAME_TRANSCRIPT = config_mod.BUCKET_NAME_FMTR_TRANSCRIPT_GOOGLE.format(
        'goog')
TRANSCRIPT_TYPE = str


def _check_for_config():
    return os.getenv('GOOGLE_APPLICATION_CREDENTIALS') is not None


class Transcriber(TranscriberBaseClass):

    SUPPORTED_FORMATS = ['flac']
    cost_per_15_seconds = [.004, .006, .009]
    no_config_error_message = (
            'Please sign up for the Google Speech-to-Text API '
            'and put the path to your credentials in an '
            'environment variable "GOOGLE_APPLICATION_CREDENTIALS"'
            )
    transcript_type = TRANSCRIPT_TYPE

    if _check_for_config():
        speech_client = speech.SpeechClient()
        storage_client = storage.Client()
        transcript_bucket = storage_client.get_bucket(BUCKET_NAME_TRANSCRIPT)

    def __init__(self, filepath):
        super().__init__(filepath)

    @classmethod
    def _setup(cls):
        super()._setup()
        if not shutil.which('gsutil'):
            raise exceptions.DependencyRequired(
                'Please install gcloud using the steps here:'
                'https://cloud.google.com/storage/docs/gsutil_install')

        cls._make_bucket_if_doesnt_exist(BUCKET_NAME_TRANSCRIPT)

    @classmethod
    def _make_bucket_if_doesnt_exist(cls, bucket_name):
        try:
            cls.storage_client.create_bucket(bucket_name)
        except gc_exceptions.Conflict:
            # this might fail if a bucket by the name exists *anywhere* on GCS?
            return
        else:
            print('made Google Cloud Storage Bucket for transcripts')

    def convert_file_format_if_needed(self):
        if self.file_format not in self.SUPPORTED_FORMATS:
            if not shutil.which('ffmpeg'):
                raise exceptions.DependencyRequired('please install ffmpeg')
            self.filepath = helpers.convert_file(self.filepath, 'flac')

    @property
    def file_format(self):
        return pathlib.Path(self.filepath).suffix[1:].lower()

    @staticmethod
    def check_for_config() -> bool:
        return _check_for_config()

    def upload_file_if_too_big(self):
        """10MB limit as of Mar 7, 2019"""
        pass

    def transcribe(self, **kwargs) -> str:
        self.convert_file_format_if_needed()
        self.upload_file_if_too_big()
        self._request_transcription(**kwargs)

    def _check_if_transcript_exists(self, transcript_name=None):
        return storage.Blob(
                    bucket=self.transcript_bucket, 
                    name=transcript_name or self.basename
                           ).exists(self.storage_client)

    def _request_transcription(
            self, 
            language_code='en-US',
            enable_automatic_punctuation=True,
            enable_speaker_diarization=True,
            num_speakers=2,
            model='phone_call',
            use_enhanced=True,
            ) -> str:
        """Returns the job_name"""
        if self._check_if_transcript_exists():
            raise exceptions.AlreadyExistsError(
                f'{self.basename} already exists on {NAME}')
        num_audio_channels = helpers.get_num_audio_channels(self.filepath)
        sample_rate = helpers.get_sample_rate(self.filepath)

        with io.open(self.filepath, 'rb') as audio_file:
            content = audio_file.read()
            audio = speech.types.RecognitionAudio(content=content)

        config = speech.types.RecognitionConfig(
            encoding=speech.enums.RecognitionConfig.AudioEncoding.FLAC,
            sample_rate_hertz=sample_rate,
            audio_channel_count=num_audio_channels,
            enable_separate_recognition_per_channel=True,
            enable_word_confidence=True,
            enable_word_time_offsets=True,
            language_code=language_code,
            enable_automatic_punctuation=enable_automatic_punctuation,
            enable_speaker_diarization=enable_speaker_diarization,
            diarization_speaker_count=num_speakers,
            model=model,
            use_enhanced=use_enhanced,
            )

        self.operation = self.speech_client.long_running_recognize(config, 
                                                                   audio)

        print('transcribing...')
        while not self.operation.done():
            sleep(1)
            print('.')

        result_list = []

        for result in self.operation.result().results:
            result_list.append(str(result))

        print('saving transcript')
        transcript_path = '/tmp/transcript.txt'
        with open(transcript_path, 'w') as fout:
            fout.write('\n'.join(result_list))
        print('uploading transcript')
        self.upload_file(BUCKET_NAME_TRANSCRIPT, transcript_path)
        os.remove(transcript_path)

        return self.basename

    @classmethod
    def retrieve_transcript(cls, transcription_job_name: str
            ) -> TRANSCRIPT_TYPE:
        """Get transcript from BUCKET_NAME_TRANSCRIPT"""
        if not cls._check_if_transcript_exists(
                cls,
                transcript_name=transcription_job_name):
            raise exceptions.DoesntExistError('no such transcript!')
        blob = cls.transcript_bucket.blob(transcription_job_name)
        f = tempfile.NamedTemporaryFile(delete=False)
        f.close()

        blob.download_to_filename(f.name)
        with open(f.name) as fin:
            transcript_text = fin.read()

        os.remove(f.name)
        return transcript_text

    def upload_file(self, bucket_name, path):
        blob = self.transcript_bucket.blob(self.basename)
        blob.upload_from_filename(path)

    @classmethod
    def get_transcription_jobs(cls, job_name_query=None, status=None) -> List[dict]:

        if status and status.lower() != 'completed':
            return []

        jobs = []

        for t in cls.transcript_bucket.list_blobs():
            if job_name_query is not None and t.name != job_name_query:
                continue
            jobs.append({'name': t.name, 'status': 'COMPLETED'})

        return jobs



import json
from pprint import pprint
import sqlite3
import sys

import click

from tatt import config, exceptions, helpers, vendors
from tatt.helpers import get_transcription_jobs, get_service


@click.group()
def cli():
    pass


@cli.command()
@click.option('-s', '--save', is_flag=True, help='save to file')
@click.option('-p', '--pretty', is_flag=True, 
              help='pretty print, will make output non-pipeable')
@click.argument('name')
def get(name, save, pretty):
    """Downloads and/or saves completed transcript."""
    try:
        transcript, service = helpers.get_transcript(name)
    except exceptions.DoesntExistError:
        raise click.ClickException(f'no such transcript {name}')
    except exceptions.NotAvailable as e:
        raise click.ClickException(str(e))

    file = None
    if service.transcript_type == dict:
        transcript = json.dumps(transcript, indent=4 if pretty else None)
        filepath = f'{name}.json'
    else:
        filepath = f'{name}.txt'

    if save:
        file = open(filepath, 'w')

    click.echo(transcript, file=file)

    if file:
        click.echo(f'Saved transcript to {filepath}.')
        file.close()



@cli.command()
@click.option('--service', type=str, help="STT service name")
@click.option('--status', type=str, help="completed | failed | in_progress")
def list(service, status):
    """Lists available STT services."""
    if service is not None and service not in vendors.SERVICES:
        raise click.ClickException(f'no such service: {service}')

    try:
        all_jobs = get_transcription_jobs(service_name=service, status=status)
    except exceptions.ConfigError as e:
        raise click.ClickException(str(e))
    else:
        if not all_jobs:
            raise click.ClickException('no transcripts currently!') 

        helpers.print_transcription_jobs(all_jobs)


@cli.command()
@click.option('-f', '--free-only', is_flag=True, help='only free services')
def services(free_only):
    """Lists available speech-to-text services."""
    click.echo(helpers.make_string_all_services(free_only))


@cli.command()
@click.argument('job_name', type=str)
def status(job_name):
    """Check the status of a transcription job."""
    try:
        jobs = get_transcription_jobs(name=job_name)
    except exceptions.DoesntExistError:
        raise click.ClickException('no job by that name')
    for job_list in jobs.values():
        for job in job_list:
            click.echo(f'{job["name"]}\t{job["status"]}')
            break
        break


@cli.command()
@click.option('--punctuation', is_flag=True, default=True, 
              help='only for Google Speech, defaults to True')
@click.option('--speaker-id/--no-speaker-id', is_flag=True, default=True, 
              help='only for google and amazon, defaults to True')
@click.option('--num_speakers', default=2, type=int,
              help='only for google and amazon, defaults to 2')
@click.option('--model', default='phone_call', 
              help='only for Google Speech, defaults to "phone_call"')
@click.option('--use-enhanced', is_flag=True, default=True,
              help='only for Google Speech, defaults to True')
@click.argument('media_filepath', type=str)
@click.argument('service_name', type=str)
def this(media_filepath, 
         service_name, 
         punctuation,
         speaker_id,
         num_speakers,
         model,
         use_enhanced):
    """Sends a media file to be transcribed."""
    if service_name == 'google':
        transcribe_kwargs = dict(
            enable_automatic_punctuation=punctuation,
            enable_speaker_diarization=speaker_id,
            model=model,
            use_enhanced=use_enhanced,
            num_speakers=num_speakers,
            )
    elif service_name == 'amazon':
        transcribe_kwargs = dict(
            enable_speaker_diarization=speaker_id,
            num_speakers=num_speakers,
                )
    else:
        transcribe_kwargs = {}
    try:
        service = get_service(service_name)
    except KeyError as e:
        raise click.ClickException(
            f'No such service! {helpers.make_string_all_services()}')

    try:
        s = service(media_filepath)
    except exceptions.ConfigError as e:
        raise click.ClickException(str(e))

    click.echo(
      f'Okay, transcribing {media_filepath} using {service_name}...')

    try:
        job_num = s.transcribe(**transcribe_kwargs)
    except exceptions.AlreadyExistsError as e:
        raise click.ClickException(str(e))
    click.echo(f'Okay, job {job_num} is being transcribed.  Use "get" '
           'command to download it.')
